from unittest.mock import patch

import pytest

import nats_app as ncli

from .mock import MockClient


@pytest.fixture
async def nc():
    with patch("nats_app.app.NATS", new=MockClient):
        _nats = ncli.NATSApp(url=["nats://127.0.0.1:4220"])
        yield _nats



import asyncio
import importlib
import logging
from collections.abc import Sequence
from enum import Enum
from typing import Any, Optional

import nats
from nats.aio import subscription
from nats.errors import BadSubscriptionError
from nats.js import JetStreamContext, api, client
from nats.js.api import KeyValueConfig, StorageType, StreamConfig
from nats.js.client import KV_STREAM_TEMPLATE
from nats.js.errors import BucketNotFoundError, NotFoundError

from nats_app.errors import default_error_handler
from nats_app.marshaling import from_bytes, normalize_payload, to_bytes
from nats_app.meta import PullSubscriptionMeta, PushSubscriptionMeta, SubscriptionMeta
from nats_app.middlewares.common import errors_handler, responder
from nats_app.router import NATSRouter
from nats_app.tasks_queue import TaskQueue

logger = logging.getLogger(__name__)

JS_DENY_CHANGE_PARAMS = ("name", "num_replicas", "sealed")

JS_ALLOWED_CHANGE_PARAMS = (
    "subjects",
    "max_consumers",
    "max_msgs",
    "max_bytes",
    "max_age",
    "storage",
    "retention",
    "discard",
    "max_msg_size",
    "duplicate_window",
    "allow_rollup_hdrs",
    "deny_delete",
    "deny_purge",
)


class NATS(nats.NATS):
    """NATS client is proxy-class with additional features."""

    def __init__(self) -> None:
        super().__init__()
        self._js_pull_subscribers_tasks = set()

    async def push_from_pull_subscribe(self, js, r: PullSubscriptionMeta):
        """Create a pull subscription handler and run it in background."""
        try:
            sub = await js.pull_subscribe(
                subject=r.subject,
                durable=r.durable,
                stream=r.stream,
                config=r.config,
                pending_msgs_limit=r.pending_msgs_limit,
                pending_bytes_limit=r.pending_bytes_limit,
                inbox_prefix=r.inbox_prefix,
            )
            info = await sub.consumer_info()
        except NotFoundError as e:
            raise ValueError(f"stream '{r.stream}' not found") from e

        pull_task = None

        async def _task():
            try:
                while True:
                    try:
                        msgs = await sub.fetch(batch=r.batch, timeout=r.timeout, heartbeat=r.heartbeat)
                        await r.handler(msgs)
                    except TimeoutError:
                        logger.debug(
                            f"subject={r.subject} stream: {info.stream_name} consumer:{info.name} pull message timeout"
                        )
                        continue
                    except Exception:
                        logger.exception(
                            f"subject={r.subject} stream: {info.stream_name} consumer:{info.name} process error"
                        )
                    await asyncio.sleep(1)
            finally:
                if pull_task:
                    self._js_pull_subscribers_tasks.remove(pull_task)
                logger.info(f"stop pull subscription on stream: {info.stream_name} consumer:{info.name}")

        pull_task = asyncio.create_task(_task())
        self._js_pull_subscribers_tasks.add(pull_task)
        logger.info(f"start pull subscription on '{r.subject}' stream: '{info.stream_name}' consumer: '{info.name}'")

    async def _js_pull_subscriber_stop(self):
        """Stop all pull subscribers background tasks."""
        for task in self._js_pull_subscribers_tasks:
            task.cancel()
        self._js_pull_subscribers_tasks = set()


class NATSApp:
    _connection_kwargs: dict
    _nc: NATS | None = None
    _js: Optional[JetStreamContext] = None
    _js_opts: dict[str, Any] = {}
    _jetstream_configs: list[StreamConfig]
    _kv_configs: list[KeyValueConfig]
    _push_subscribers: list[SubscriptionMeta]
    _js_push_subscribers: list[PushSubscriptionMeta]
    _subscriptions: list[subscription.Subscription] = []
    _js_pull_subscribers_tasks: set[asyncio.Task] = set()
    middlewares: Optional[list]

    def __init__(
        self,
        servers: Optional[list[str]] = None,
        js_opts: Optional[dict[str, Any]] = None,
        middlewares: Optional[Sequence] = None,
        **kwargs,
    ) -> None:
        """
        Create a new NATSApp instance.
        """
        self._connection_kwargs = kwargs
        if servers:
            self._connection_kwargs["servers"] = servers

        self.middlewares = middlewares or []
        self._jetstream_configs = []
        self._kv_configs = []
        self._push_subscribers = []
        self._js_push_subscribers = []
        self._js_pull_subscribers = []
        self._subscriptions = []
        self._js_opts = js_opts or {}
        self._task_queues = []

        def _error_handler(m, e):
            return self._error_handler(m, e)

        self.middlewares = [
            responder(),
            errors_handler(_error_handler),
        ]
        self.set_error_handler(default_error_handler)

    async def connect(self, **options):
        """Connect to NATS server."""

        async def disconnected_cb():
            logger.info("NATS disconnected")
            await self._js_pull_subscriber_stop()

        async def reconnected_cb():
            logger.info(f"NATS reconnected {self._nc.connected_url.netloc}")
            await self._js_pull_subscribe_all()
            logger.info("NATS resubscribe")

        async def error_cb(e):
            logger.info(f"NATS was an error: {e}")

        async def closed_cb():
            logger.info("NATS connection is closed")

        self._nc = NATS()

        connection_kwargs = self._connection_kwargs.copy()
        connection_kwargs.update(
            {
                "disconnected_cb": disconnected_cb,
                "reconnected_cb": reconnected_cb,
                "error_cb": error_cb,
                "closed_cb": closed_cb,
            }
        )
        connection_kwargs.update(options)

        await self._nc.connect(**connection_kwargs)
        logger.info(f"Connected to NATS on {self._nc.connected_url.netloc}")
        self._bind_task_queue()
        await self._streams_create_or_update()
        await self._kv_create_or_update()
        await self.subscribe_all()

    def __getattr__(self, name):
        return getattr(self._nc, name)

    @property
    def js(self):
        """Return JetStreamContext instance."""
        if not self._js:
            self._js = self._nc.jetstream(**self._js_opts)
        return self._js

    async def unsubscribe_all(self):
        """Unsubscribe all subscriptions."""
        await self._js_pull_subscriber_stop()
        if self._nc and not self._nc.is_closed:
            for subscriber in self._subscriptions:
                try:
                    if not subscriber._closed:
                        await subscriber.drain()
                    if not subscriber._closed:
                        await subscriber.unsubscribe()
                except BadSubscriptionError as e:
                    raise ValueError(f"nats: invalid subscription: {subscriber.subject}") from e
        self._subscriptions = []

    @classmethod
    def _stream_config_dict(cls, config: StreamConfig) -> dict:
        """Convert StreamConfig to dict with default values."""
        new = {k: v if not isinstance(v, Enum) else v.value for k, v in config.as_dict().items()}
        if new.get("duplicate_window") == 0:
            new["duplicate_window"] = 120000000000  # default value
        return new

    @classmethod
    def _is_equal(cls, exist, new, fields) -> bool:
        """Check if two config are equal."""
        return all([exist.get(n) == new.get(n) for n in fields if new.get(n) is not None])

    @classmethod
    def _get_change_dict(cls, exist, new, fields):
        """Return a dict with changed"""
        return {
            n: {"old": exist.get(n), "new": new.get(n)}
            for n in fields
            if new.get(n) is not None and exist.get(n) != new.get(n)
        }

    async def _streams_create_or_update(self):
        """Create or update JetStream streams."""
        for config in self._jetstream_configs:
            if config.name is None:
                raise ValueError("nats: stream name is required")
            try:
                si = await self.js.stream_info(config.name)
                exist = self._stream_config_dict(si.config)
                new = self._stream_config_dict(config)

                if not self._is_equal(exist, new, JS_DENY_CHANGE_PARAMS):
                    change = self._get_change_dict(exist, new, JS_DENY_CHANGE_PARAMS)
                    logger.error(f"deny update stream {change}")
                    raise ValueError(f"nats: stream config params {JS_DENY_CHANGE_PARAMS} deny change")

                if not self._is_equal(exist, new, JS_ALLOWED_CHANGE_PARAMS):
                    change = self._get_change_dict(exist, new, JS_ALLOWED_CHANGE_PARAMS)
                    si = await self.js.update_stream(config)
                    logger.info(f"update stream {change} after stream info: {si.as_dict()}")
                else:
                    logger.info(f"unchanged stream: {config.name}")

            except NotFoundError:
                si = await self.js.add_stream(config)
                logger.info(f"add stream info: {si.as_dict()}")

    async def _kv_create_or_update(self):
        """Create or update JetStream streams."""
        for config in self._kv_configs:
            if config.bucket is None:
                raise ValueError("nats: kv bucket name is required")
            try:
                kv = await self.js.key_value(config.bucket)
                status = await kv.status()
                si = status.stream_info

                if si.config.storage != config.storage:
                    logger.warning(f"can not change storage of kv bucket: {status.bucket}")
                if si.config.republish != config.republish:
                    logger.warning(f"can not change republish of kv bucket: {status.bucket}")

                if (
                    si.config.max_bytes != config.max_bytes
                    or si.config.max_age != config.ttl
                    or si.config.replicas != config.replicas
                    or si.config.placement != config.placement
                ):
                    assert si.config.name == KV_STREAM_TEMPLATE.format(bucket=config.bucket)
                    si.config.max_bytes = config.max_bytes
                    si.config.max_age = config.ttl
                    si.config.replicas = config.replicas
                    si.config.placement = config.placement
                    await self.js.update_stream(si.config)
                    logger.info(f"update kv bucket: {status.bucket}")
                else:
                    logger.info(f"unchanged kv bucket: {status.bucket}")
            except BucketNotFoundError:
                kv = await self.js.create_key_value(config)
                status = await kv.status()
                logger.info(f"create kv bucket: {status.bucket}")

    async def _js_pull_subscribe_all(self):
        """Subscribe all jetstream pull subscribers."""
        for r in self._js_pull_subscribers:
            await self._register_js_pull_subscriber(r)

    async def subscribe_all(self):
        """Subscribe all subscribers."""
        for r in self._push_subscribers:
            await self._register_handler(r)

        for r in self._js_push_subscribers:
            await self._register_js_push_subscriber(r)

        await self._js_pull_subscribe_all()

    async def close(self):
        """Close NATS connection."""
        await self.unsubscribe_all()

        if self._nc and not self._nc.is_closed:
            await self._nc.drain()
        if self._nc and not self._nc.is_closed:
            await self._nc.flush()
        if self._nc and not self._nc.is_closed:
            await self._nc.close()
        self._js = None

    def set_error_handler(self, fn):
        """Set error handler."""
        self._error_handler = fn  # noqa

    def add_middlewares(self, middlewares):
        """Add NATS App middlewares."""
        for m in middlewares:
            if isinstance(m, str):
                try:
                    mod, name = m.rsplit(".", 1)
                    mod = importlib.import_module(mod)
                    m = getattr(mod, name)
                except ModuleNotFoundError as e:
                    raise ModuleNotFoundError(f"Module {m} not found") from e
            self.middlewares.append(m)

    async def publish(
        self,
        subject: str,
        payload: Optional[dict] | bytes = None,
        reply: str = "",
        headers: Optional[dict[str, str]] = None,
    ) -> None:
        """Publish a message."""
        if not isinstance(payload, bytes):
            payload = normalize_payload(payload)
            payload = to_bytes(payload)
        await self._nc.publish(subject, payload, reply, headers)

    async def request(
        self,
        subject: str,
        payload: Any,
        timeout: float = 0.5,
        old_style: bool = False,
        headers: Optional[dict[str, Any]] = None,
    ):
        """Send a request."""
        if not isinstance(payload, bytes):
            payload = normalize_payload(payload)
            payload = to_bytes(payload)
        result = await self._nc.request(subject, payload, timeout, old_style, headers)
        if result is not None:
            result = from_bytes(result.data)
        return result

    def push_subscribe(
        self,
        subject,
        queue: str = "",
        future: Optional[asyncio.Future] = None,
        max_msgs: int = 0,
        pending_msgs_limit: int = subscription.DEFAULT_SUB_PENDING_MSGS_LIMIT,
        pending_bytes_limit: int = subscription.DEFAULT_SUB_PENDING_BYTES_LIMIT,
    ):
        """Decorator for push_subscribe handler."""

        def wrapper(fn):
            self._push_subscribers.append(
                SubscriptionMeta(
                    subject=subject,
                    queue=queue,
                    handler=fn,
                    future=future,
                    max_msgs=max_msgs,
                    pending_msgs_limit=pending_msgs_limit,
                    pending_bytes_limit=pending_bytes_limit,
                )
            )

        return wrapper

    def js_push_subscribe(
        self,
        subject: str,
        queue: Optional[str] = None,
        durable: Optional[str] = None,
        stream: Optional[str] = None,
        config: Optional[api.ConsumerConfig] = None,
        manual_ack: bool = False,
        ordered_consumer: bool = False,
        idle_heartbeat: Optional[float] = None,
        flow_control: bool = False,
        pending_msgs_limit: int = client.DEFAULT_JS_SUB_PENDING_MSGS_LIMIT,
        pending_bytes_limit: int = client.DEFAULT_JS_SUB_PENDING_BYTES_LIMIT,
        deliver_policy: Optional[api.DeliverPolicy] = None,
        headers_only: Optional[bool] = None,
        inactive_threshold: Optional[float] = None,
    ):
        """Decorator for jetstream push_subscribe handler."""

        def wrapper(fn):
            self._js_push_subscribers.append(
                PushSubscriptionMeta(
                    subject=subject,
                    queue=queue,
                    cb=fn,
                    durable=durable,
                    stream=stream,
                    config=config,
                    manual_ack=manual_ack,
                    ordered_consumer=ordered_consumer,
                    idle_heartbeat=idle_heartbeat,
                    flow_control=flow_control,
                    pending_msgs_limit=pending_msgs_limit,
                    pending_bytes_limit=pending_bytes_limit,
                    deliver_policy=deliver_policy,
                    headers_only=headers_only,
                    inactive_threshold=inactive_threshold,
                )
            )

        return wrapper

    def js_pull_subscribe(
        self,
        subject: str,
        durable: Optional[str] = None,
        stream: Optional[str] = None,
        config: Optional[api.ConsumerConfig] = None,
        pending_msgs_limit: int = client.DEFAULT_JS_SUB_PENDING_MSGS_LIMIT,
        pending_bytes_limit: int = client.DEFAULT_JS_SUB_PENDING_BYTES_LIMIT,
        inbox_prefix: bytes = api.INBOX_PREFIX,
        batch: int = 1,
        timeout: Optional[float] = None,
        heartbeat: Optional[float] = None,
    ):
        """Decorator for jetstream pull_subscribe handler."""

        def wrapper(fn):
            self._js_pull_subscribers.append(
                PullSubscriptionMeta(
                    handler=fn,
                    subject=subject,
                    durable=durable,
                    stream=stream,
                    config=config,
                    pending_msgs_limit=pending_msgs_limit,
                    pending_bytes_limit=pending_bytes_limit,
                    inbox_prefix=inbox_prefix,
                    batch=batch,
                    timeout=timeout,
                    heartbeat=heartbeat,
                )
            )

        return wrapper

    @property
    def register_routers(self):
        """Register NATSRouter instances."""

        def _fn(obj):
            if isinstance(obj, NATSRouter):
                # push_subscribe RPC handlers
                for r in obj.push_subscribers:
                    self._push_subscribers.append(r)

                for r in obj.js_push_subscribers:
                    self._js_push_subscribers.append(r)

                for r in obj.js_pull_subscribers:
                    self._js_pull_subscribers.append(r)

            elif isinstance(obj, StreamConfig):
                self._jetstream_configs.append(obj)

            elif isinstance(obj, KeyValueConfig):
                self._kv_configs.append(obj)

        return _fn

    async def _register_handler(self, r: SubscriptionMeta):
        """Register push_subscribe handler."""
        handler = r.handler
        for m in reversed(self.middlewares):
            handler = m(handler)

        sub = await self._nc.subscribe(
            r.subject,
            queue=r.queue,
            cb=handler,
            future=r.future,
            max_msgs=r.max_msgs,
            pending_msgs_limit=r.pending_msgs_limit,
            pending_bytes_limit=r.pending_bytes_limit,
        )
        self._subscriptions.append(sub)
        logger.info("NATS push_subscribe handler on subject: %s", r.subject)

    async def _register_js_push_subscriber(self, r: PushSubscriptionMeta):
        """Register jetstream push_subscribe handler."""
        try_connect = 3  # wait nats jetstream started
        while True:
            try:
                sub = await self.js.subscribe(
                    subject=r.subject,
                    queue=r.queue,
                    cb=r.cb,
                    durable=r.durable,
                    stream=r.stream,
                    config=r.config,
                    manual_ack=r.manual_ack,
                    ordered_consumer=r.ordered_consumer,
                    idle_heartbeat=r.idle_heartbeat,
                    flow_control=r.flow_control,
                    pending_msgs_limit=r.pending_msgs_limit,
                    pending_bytes_limit=r.pending_bytes_limit,
                    deliver_policy=r.deliver_policy,
                    headers_only=r.headers_only,
                    inactive_threshold=r.inactive_threshold,
                )
                self._subscriptions.append(sub)
                logger.info("NATS push_subscribe jetstream push handler on subject: %s", r.subject)
                return
            except asyncio.TimeoutError:
                try_connect -= 1
                if try_connect == 0:
                    raise
                logger.info("NATS Connect to nats jetstream timeout. Retry connect.")

    async def _register_js_pull_subscriber(self, r: PullSubscriptionMeta):
        """Register jetstream pull_subscribe handler."""
        if self._nc:
            await self._nc.push_from_pull_subscribe(self.js, r)

    async def _js_pull_subscriber_stop(self):
        """Stop all jetstream pull subscribers."""
        if self._nc:
            await self._nc._js_pull_subscriber_stop()

    def _bind_task_queue(self):
        """Bind task queue."""
        for tq in self._task_queues:
            tq.bind(self)

    def create_queue(
        self,
        subjects: list[str],
        stream_name: Optional[str] = None,
        storage: StorageType = StorageType.FILE,
        stream_config: Optional[StreamConfig] = None,
        durable: Optional[str] = None,
    ) -> TaskQueue:
        tq = TaskQueue(
            subjects=subjects,
            stream_name=stream_name,
            storage=storage,
            stream_config=stream_config,
            durable=durable,
        )
        self._task_queues.append(tq)
        return tq

    def register_task_queue(self, *tasks_queues: TaskQueue):
        """Register task queue."""
        for tq in tasks_queues:
            self._task_queues.append(tq)



import uuid
from typing import Any

import ujson
from pydantic import BaseModel


def to_bytes(data: Any) -> bytes:
    if isinstance(data, bytes):
        return data
    return ujson.dumps(data).encode("utf-8")


def from_bytes(data: bytes) -> Any:
    return ujson.loads(data.decode("utf-8")) if data else {}


def normalize_payload(data: Any) -> Any:
    if isinstance(data, bytes):
        return data
    if isinstance(data, BaseModel):
        return data.model_dump(mode="json", exclude_none=True)
    if isinstance(data, (dict,)):
        return {k: normalize_payload(v) for k, v in data.items() if v is not None}
    if isinstance(data, (list, tuple)):
        return [normalize_payload(r) for r in data]
    if isinstance(data, uuid.UUID):
        return str(data)
    return data



from functools import wraps
from typing import Optional

from nats.aio.msg import Msg
from pydantic import validate_call

from nats_app.marshaling import from_bytes, normalize_payload, to_bytes


def bytes_to_str(by_pass_msg: Optional[str] = None):
    _by_pass_msg = None

    def _func(next_handler):
        @wraps(next_handler)
        async def wrap(msg: Msg):
            data = from_bytes(msg.data)
            kwargs = {}
            if by_pass_msg:
                kwargs[by_pass_msg] = msg
            result = await next_handler(data, **kwargs)
            if result is not None and not isinstance(result, bytes):
                result = to_bytes(normalize_payload(result))
            return result

        return wrap

    if not callable(by_pass_msg):
        _by_pass_msg = by_pass_msg
        return _func
    return _func(by_pass_msg)


def validate_args(by_pass_msg: Optional[str] = None):
    _by_pass_msg = None

    def _func(next_handler):
        @wraps(next_handler)
        async def wrap(msg: Msg):
            data = from_bytes(msg.data)
            args, kwargs = data.get("args", []), data.get("kwargs", {})
            if _by_pass_msg:
                kwargs[_by_pass_msg] = msg
            result = await validate_call(next_handler)(*args, **kwargs)
            if result is not None and not isinstance(result, bytes):
                result = to_bytes(normalize_payload(result))
            return result

        return wrap

    if not callable(by_pass_msg):
        _by_pass_msg = by_pass_msg
        return _func
    return _func(by_pass_msg)



import asyncio
from collections import defaultdict
from typing import Any, Optional

import pytest
from nats.aio.msg import Msg
from nats.js.api import StorageType, StreamConfig

from nats_app.app import NATSApp
from nats_app.middlewares.validation import validate_args
from nats_app.tasks_queue import TaskQueue


async def test_rpc_call(nc):
    @nc.push_subscribe("app.subject.echo", queue="worker")
    async def rpc_func_echo(m: Msg) -> str:
        return f"Response: {m.data}"

    await nc.connect()

    res = await nc.request("app.subject.echo", "test")
    assert isinstance(res, str)
    assert res == "Response: b'\"test\"'"


async def test_rpc_call_w_validate(nc):
    @nc.push_subscribe("app.subject.validate_echo", queue="worker")
    @validate_args(by_pass_msg="msg")
    async def rpc_func_echo(data: str, msg: Any) -> str:
        return f"Response: {data} Msg.data: {msg.data}"

    await nc.connect()

    res = await nc.request("app.subject.validate_echo", {"args": ["test"]})
    assert isinstance(res, str)
    assert res == 'Response: test Msg.data: b\'{"args":["test"]}\''


async def test_rpc_call_w_validate_wo_param(nc):
    @nc.push_subscribe("app.subject.validate_echo", queue="worker")
    @validate_args
    async def rpc_func_echo(data: str) -> str:
        return f"Response: {data}"

    await nc.connect()

    res = await nc.request("app.subject.validate_echo", {"args": ["test"]})
    assert isinstance(res, str)
    assert res == "Response: test"


async def test_rpc_call_invalid_args(nc):
    @nc.push_subscribe("app.subject.validate_echo", queue="worker")
    @validate_args(by_pass_msg="msg")
    async def rpc_func_echo(data: str, val: int, key: bool, var1: Optional[float], var2: float, msg: Any) -> str:
        return f"Response: {data} Msg.data: {msg.data}"

    await nc.connect()

    res = await nc.request(
        "app.subject.validate_echo",
        {"args": ["test", "wow"], "kwargs": {"var1": "oto", "var2": 3.14}},
    )
    assert isinstance(res, dict)
    assert res == {
        "errors": [
            {
                "loc": [1],
                "msg": "Input should be a valid integer, unable to parse string as an integer",
                "type": "int_parsing",
            },
            {
                "loc": ["key"],
                "msg": "Missing required argument",
                "type": "missing_argument",
            },
            {
                "loc": ["var1"],
                "msg": "Input should be a valid number, unable to parse string as a number",
                "type": "float_parsing",
            },
        ]
    }


async def test_publish_w_validate(nc):
    result = {}

    @nc.push_subscribe("app.subject.pub")
    @validate_args()
    async def rpc_func_echo(data: str, data1: int) -> None:
        result["data"] = data
        result["data1"] = data1

    await nc.connect()

    await nc.publish("app.subject.pub", {"args": ["test"], "kwargs": {"data1": 1}})
    assert result["data"] == "test"
    assert result["data1"] == 1


async def test_publish_wo_validate(nc):
    result = {}

    @nc.push_subscribe("app.subject.pub_raw")
    async def rpc_func_echo(m: Msg) -> None:
        result["data"] = m.data

    await nc.connect()

    await nc.publish("app.subject.pub_raw", {"args": ["test"], "kwargs": {"data1": 1}})
    assert result["data"] == b'{"args":["test"],"kwargs":{"data1":1}}'


async def test_publish_w_queue(nc):
    c = defaultdict(int)
    result = {}

    @nc.push_subscribe("app.subject.pub_raw", queue="worker")
    async def rpc_func_echo(m: Msg) -> None:
        c["call"] += 1
        result["data"] = m.data

    @nc.push_subscribe("app.subject.pub_raw", queue="worker")
    async def rpc_func_echo_v2(m: Msg) -> None:
        c["call"] += 1
        result["data"] = m.data

    await nc.connect()

    await nc.publish("app.subject.pub_raw", {"args": ["test"], "kwargs": {"data1": 1}})
    assert result["data"] == b'{"args":["test"],"kwargs":{"data1":1}}'
    assert c["call"] == 1


async def test_publish_wo_queue(nc):
    c = defaultdict(int)
    result = {}

    @nc.push_subscribe("app.subject.pub_raw")
    async def rpc_func_echo(m: Msg) -> None:
        c["call"] += 1
        result["data"] = m.data

    @nc.push_subscribe("app.subject.pub_raw")
    async def rpc_func_echo_v2(m: Msg) -> None:
        c["call"] += 1
        result["data"] = m.data

    await nc.connect()

    await nc.publish("app.subject.pub_raw", {"args": ["test"], "kwargs": {"data1": 1}})
    assert result["data"] == b'{"args":["test"],"kwargs":{"data1":1}}'
    assert c["call"] == 2


async def test_push_subscription(nc):
    result = None

    @nc.js_push_subscribe("app.subject.validate_echo", queue="worker")
    async def handler(msg: Msg):
        nonlocal result
        result = msg.data
        await msg.ack()

    await nc.connect()

    await nc.js.publish("app.subject.validate_echo", b"TEST123")
    assert isinstance(result, bytes)
    assert result == b"TEST123"


async def test_pull_subscription(nc):
    result = None

    @nc.js_pull_subscribe("app.subject.validate_echo", batch=1)
    async def handler(msgs: list[Msg]):
        assert isinstance(msgs, list)
        assert len(msgs) == 1
        nonlocal result
        result = msgs[0].data

    await nc.connect()
    await nc.js.publish("app.subject.validate_echo", b"TEST123")
    assert isinstance(result, bytes)
    assert result == b"TEST123"


async def test_pull_subscription_multihandlers_wo_group(nc):
    result1 = None
    result2 = None

    @nc.js_pull_subscribe("app.subject.validate_echo", batch=1)
    async def handler(msgs: list[Msg]):
        assert isinstance(msgs, list)
        assert len(msgs) == 1
        nonlocal result1
        result1 = msgs[0].data

    @nc.js_pull_subscribe("app.subject.validate_echo", batch=1)
    async def handler2(msgs: list[Msg]):
        assert isinstance(msgs, list)
        assert len(msgs) == 1
        nonlocal result2
        result2 = msgs[0].data

    await nc.connect()
    await nc.js.publish("app.subject.validate_echo", b"TEST123")
    assert isinstance(result1, bytes)
    assert result1 == b"TEST123"
    assert isinstance(result2, bytes)
    assert result2 == b"TEST123"


async def test_pull_subscription_multihandlers_w_one_group(nc):
    result1 = None
    result2 = None

    @nc.js_pull_subscribe("app.subject.validate_echo", durable="group1", batch=1)
    async def handler(msgs: list[Msg]):
        assert isinstance(msgs, list)
        assert len(msgs) == 1
        nonlocal result1
        result1 = msgs[0].data

    @nc.js_pull_subscribe("app.subject.validate_echo", durable="group1", batch=1)
    async def handler2(msgs: list[Msg]):
        assert isinstance(msgs, list)
        assert len(msgs) == 1
        nonlocal result2
        result2 = msgs[0].data

    await nc.connect()
    await nc.js.publish("app.subject.validate_echo", b"TEST123")
    assert (result1 is None and result2 is not None) or (result1 is not None and result2 is None)
    if result1 is not None:
        assert isinstance(result1, bytes)
        assert result1 == b"TEST123"
    else:
        assert result1 is None
        assert isinstance(result2, bytes)
        assert result2 == b"TEST123"


async def test_pull_subscription_multihandlers_w_different_group(nc):
    result1 = None
    result2 = None

    @nc.js_pull_subscribe("app.subject.validate_echo", durable="group1", batch=1)
    async def handler(msgs: list[Msg]):
        assert isinstance(msgs, list)
        assert len(msgs) == 1
        nonlocal result1
        result1 = msgs[0].data

    @nc.js_pull_subscribe("app.subject.validate_echo", durable="group2", batch=1)
    async def handler2(msgs: list[Msg]):
        assert isinstance(msgs, list)
        assert len(msgs) == 1
        nonlocal result2
        result2 = msgs[0].data

    await nc.connect()
    await nc.js.publish("app.subject.validate_echo", b"TEST123")
    assert isinstance(result1, bytes)
    assert result1 == b"TEST123"
    assert isinstance(result2, bytes)
    assert result2 == b"TEST123"


@pytest.mark.asyncio
async def test_create_queue_basic():
    app = NATSApp()
    subjects = ["subject1", "subject2"]
    queue = app.create_queue(subjects=subjects)

    assert isinstance(queue, TaskQueue)
    assert queue.subjects == subjects
    assert queue.stream_name is None
    assert queue.storage == StorageType.FILE
    assert queue.durable is None
    assert queue in app._task_queues


@pytest.mark.asyncio
async def test_create_queue_with_stream_name():
    app = NATSApp()
    subjects = ["subject1"]
    stream_name = "test_stream"
    queue = app.create_queue(subjects=subjects, stream_name=stream_name)

    assert isinstance(queue, TaskQueue)
    assert queue.subjects == subjects
    assert queue.stream_name == stream_name
    assert queue.storage == StorageType.FILE
    assert queue.durable is None
    assert queue in app._task_queues


@pytest.mark.asyncio
async def test_create_queue_with_custom_storage():
    app = NATSApp()
    subjects = ["subject1"]
    storage = StorageType.MEMORY
    queue = app.create_queue(subjects=subjects, storage=storage)

    assert isinstance(queue, TaskQueue)
    assert queue.subjects == subjects
    assert queue.stream_name is None
    assert queue.storage == storage
    assert queue.durable is None
    assert queue in app._task_queues


@pytest.mark.asyncio
async def test_create_queue_with_stream_config():
    app = NATSApp()
    subjects = ["subject1"]
    stream_config = StreamConfig(name="test_stream", subjects=["subject2"], storage=StorageType.FILE)
    queue = app.create_queue(subjects=subjects, stream_config=stream_config)

    assert isinstance(queue, TaskQueue)
    assert queue.subjects == stream_config.subjects
    assert queue.stream_name is stream_config.name
    assert queue.storage == StorageType.FILE
    assert queue.stream_config == stream_config
    assert queue.durable is None
    assert queue in app._task_queues


@pytest.mark.asyncio
async def test_create_queue_with_durable():
    app = NATSApp()
    subjects = ["subject1"]
    durable = "durable_name"
    queue = app.create_queue(subjects=subjects, durable=durable)

    assert isinstance(queue, TaskQueue)
    assert queue.subjects == subjects
    assert queue.stream_name is None
    assert queue.storage == StorageType.FILE
    assert queue.durable == durable
    assert queue in app._task_queues


@pytest.mark.asyncio
def test_register_task_queue_single():
    app = NATSApp()
    task_queue = TaskQueue(subjects=["subject1"])

    app.register_task_queue(task_queue)

    assert len(app._task_queues) == 1
    assert app._task_queues[0] == task_queue


@pytest.mark.asyncio
def test_register_task_queue_multiple():
    app = NATSApp()
    task_queue1 = TaskQueue(subjects=["subject1"])
    task_queue2 = TaskQueue(subjects=["subject2"])

    app.register_task_queue(task_queue1, task_queue2)

    assert len(app._task_queues) == 2
    assert task_queue1 in app._task_queues
    assert task_queue2 in app._task_queues


@pytest.mark.asyncio
def test_register_task_queue_no_duplicates():
    app = NATSApp()
    task_queue = TaskQueue(subjects=["subject1"])

    app.register_task_queue(task_queue)
    app.register_task_queue(task_queue)  # Register the same queue again

    assert len(app._task_queues) == 2  # Ensure duplicates are allowed
    assert app._task_queues.count(task_queue) == 2



import contextlib
import logging
from collections import defaultdict
from collections.abc import Awaitable
from functools import wraps
from typing import Callable, Optional

import ujson
from nats.aio.msg import Msg, MsgAlreadyAckdError
from nats.js.api import AckPolicy, ConsumerConfig, StorageType, StreamConfig
from pydantic import BaseModel, PrivateAttr, validate_call

from nats_app.marshaling import normalize_payload

logger = logging.getLogger(__name__)


class MetaTask(BaseModel):
    task: str | None = None
    args: tuple | list
    kwargs: dict
    _msg: Optional[any] = PrivateAttr()

    @classmethod
    def from_msg(cls, msg: Msg) -> "MetaTask":
        m = cls(**ujson.loads(msg.data))
        m._msg = msg
        return m

    async def ack(self) -> None:
        with contextlib.suppress(MsgAlreadyAckdError):
            await self._msg.ack()

    async def nak(self, delay: int) -> None:
        with contextlib.suppress(MsgAlreadyAckdError):
            await self._msg.nak(delay=delay)

    async def retry(self, delay: int, max_retry: int) -> bool:
        num_delivered = self.metadata.num_delivered
        if num_delivered > max_retry:
            await self.ack()
            return False
        else:
            delay = delay + (num_delivered - 1) * 2
            await self.nak(delay=self.retry_delay(delay) or delay)
            return True

    def retry_delay(self, delay: int) -> Optional[int]:
        return delay + (self.metadata.num_delivered - 1) * 2 if self._msg else None

    @property
    def header(self) -> Optional[dict[str, str]]:
        return self._msg.header if self._msg else None

    @property
    def metadata(self) -> Msg.Metadata:
        return self._msg.metadata if self._msg else None

    @property
    def is_acked(self) -> bool:
        return self._msg.is_acked if self._msg else None

    @property
    def sid(self) -> int:
        return self._msg.sid if self._msg else None


class TaskParams(BaseModel):
    fn: Callable[..., Awaitable]
    skip_validate: bool = False
    subject: Optional[str] = None
    batch: int = 1
    timeout: Optional[int] = None
    consumer_config: Optional[ConsumerConfig] = None
    max_retry: int = 5
    fail_delay: int = 60


def _get_fn_name(fn):
    return f"{fn.__module__}.{fn.__qualname__}"


class TaskQueue:
    def __init__(
        self,
        *,
        subjects: list[str],
        stream_name: Optional[str] = None,
        storage: StorageType = StorageType.FILE,
        stream_config: Optional[StreamConfig] = None,
        durable: Optional[str] = None,
        clean_bad_messages: bool = False,
    ):
        self._nc = None
        self._registered_tasks = {}

        self.stream_config = stream_config
        if not self.stream_config:
            self.stream_config = StreamConfig(
                name=stream_name,
                subjects=subjects,
                storage=storage,
            )

        self.durable = durable
        self.clean_bad_messages = clean_bad_messages

    @property
    def stream_name(self):
        return self.stream_config.name

    @property
    def storage(self):
        return self.stream_config.storage

    @property
    def subjects(self):
        return self.stream_config.subjects

    @property
    def default_subject(self):
        subject = self.subjects[0]
        if not subject.endswith(".*"):
            return subject
        else:
            return subject[:-1] + "task"

    def _get_subject(self, subject: Optional[str]):
        if not subject:
            return self.default_subject
        subj = subject.rstrip("*")
        for s in self.subjects:
            if subj.startswith(s.rstrip("*")):
                return subject
        raise ValueError(f"subject '{subject}' not configures in stream {self.stream_name}: subjects: {self.subjects}")

    def _set_delay(self, task_name: str, subject: str):
        async def _delay_fn(*args, **kwargs) -> None:
            if not self._nc:
                logger.error("TaskQueue: nats connection is not configured")
                raise ValueError("TaskQueue: nats connection is not configured")

            meta = MetaTask(
                task=task_name,
                args=[normalize_payload(v) for v in args],
                kwargs={k: normalize_payload(v) for k, v in kwargs.items()},
            )

            await self._nc.publish(subject, meta.model_dump(mode="json"))
            logger.info(
                f"pub task message '{meta.task}' to subject '{subject}' to stream '{self.stream_name}'"
                f"with args={meta.args} kwargs={meta.kwargs}"
            )

        return _delay_fn

    def task(
        self,
        cb=None,
        subject: Optional[str] = None,
        *,
        batch: int = 1,
        timeout: Optional[int] = None,
        ack_policy: AckPolicy = AckPolicy.EXPLICIT,
        max_retry: int = 5,
        fail_delay: int = 60,
        consumer_config: Optional[ConsumerConfig] = None,
        skip_validate: bool = False,
    ):
        def wrapper(fn):
            if not skip_validate and batch == 1:
                fn = validate_call(fn)

            fn.task_name = _get_fn_name(fn)
            fn.subject = self._get_subject(subject)
            fn.delay = self._set_delay(fn.task_name, fn.subject)

            self._registered_tasks[fn.task_name] = TaskParams(
                fn=fn,
                skip_validate=skip_validate,
                subject=fn.subject,
                batch=batch,
                timeout=timeout,
                consumer_config=consumer_config if consumer_config else ConsumerConfig(ack_policy=ack_policy),
                max_retry=max_retry,
                fail_delay=fail_delay,
            )

            logger.info(f"register delayed task: {fn.task_name}")

            @wraps(fn)
            async def wrap(*args, **kwargs):
                return await fn(*args, **kwargs)

            return wrap

        if callable(cb):
            return wrapper(cb)
        return wrapper

    def bind(self, nc):
        self._nc = nc
        nc._jetstream_configs.append(self.stream_config)

        # prepare stream subscribes
        message_tasks = defaultdict(dict)
        for task_param in self._registered_tasks.values():
            message_tasks[task_param.subject][task_param.fn.task_name] = task_param

        # subscribe to stream message
        for subject, task_params in message_tasks.items():
            self._is_valid_multi_params(subject, task_params)
            self._subscribe_on_jetstream(subject, task_params)

    @classmethod
    def _is_valid_multi_params(cls, subject: str, task_params: dict[str, TaskParams]):
        if len(task_params) == 1:
            return
        data = defaultdict(list)
        for tp in task_params.values():
            for k, v in tp.model_dump().items():
                if k == "consumer_config":
                    v = hash(frozenset(v.items()))
                if k in (
                    "batch",
                    "timeout",
                    "consumer_config",
                    "max_retry",
                    "fail_delay",
                ):
                    data[k].append(v)

        for v in data.values():
            if len(set(v)) > 1:
                raise ValueError(f"batch message on subject='{subject}' received multiple task params={task_params}")

    async def _parse_msg(self, m: Msg) -> Optional[MetaTask]:
        try:
            return MetaTask.from_msg(m)
        except Exception:
            if self.clean_bad_messages:
                logger.warning(f"fail to parse message: {m.data} - cleaning up")
                await m.ack()
            return None

    def _subscribe_on_jetstream(self, subject: str, task_params: dict[str, TaskParams]):
        tp = list(task_params.values())[0]
        is_batch = tp.batch is not None and tp.batch > 1

        @self._nc.js_pull_subscribe(
            subject,
            stream=self.stream_name,
            durable=self.durable,
            batch=tp.batch,
            timeout=tp.timeout,
            config=tp.consumer_config,
        )
        async def subscribe(msgs: list[Msg]):
            if not msgs:
                return

            data = [d for d in [await self._parse_msg(msg) for msg in msgs] if d is not None]
            if not data:
                return

            task_names = set([m.task for m in data])
            if len(task_names) > 1:
                logger.error(f"batch message received multiple task names: {task_names}")
                return

            task_name = data[0].task

            if task_name:
                task_param = task_params.get(task_name)
            elif len(task_params) == 1:
                task_param = tp
            else:
                task_param = None

            if not task_param:
                logger.error(f"error: task params not for found for message: {data}")
                return

            params = ujson.dumps([{"args": d.args, "kwargs": d.kwargs} for d in data])
            try:
                logger.info(f"call task='{task_name}'({params})")
                if is_batch:
                    result = await task_param.fn(data)
                else:
                    result = await task_param.fn(*data[0].args, **data[0].kwargs)
                logger.info(f"finish task='{task_name}'({params}) result={result}")
                if not is_batch or tp.consumer_config.ack_policy == AckPolicy.ALL:
                    await msgs[len(msgs) - 1].ack()

            except Exception:
                logger.exception(f"fail task: {task_name}({params})")
                if not is_batch:
                    msg = data[0]
                    num_delivered = msg.metadata.num_delivered
                    if num_delivered > task_param.max_retry:
                        await msg.ack()
                        logger.warning(f"stop retry fail task: {task_name}({params}) - reach max retry count")
                    else:
                        delay = task_param.fail_delay + (num_delivered - 1) * 2
                        await msg.nak(delay=delay)
                        logger.warning(f"retry fail task: {task_name}({params}) - retry={num_delivered} delay={delay}")

                else:
                    if tp.consumer_config.ack_policy == AckPolicy.ALL:
                        num_delivered = min([d.metadata.num_delivered for d in data])
                        if num_delivered > task_param.max_retry:
                            await msgs[len(msgs) - 1].ack()
                            logger.warning(f"stop retry fail task: {task_name}({params}) - reach max retry count")
                        else:
                            delay = task_param.fail_delay + (num_delivered - 1) * 2
                            await msgs[len(msgs) - 1].nak(delay=delay)
                            logger.warning(
                                f"retry fail task: {task_name}({params}) - retry={num_delivered} delay={delay}"
                            )

        logger.info(
            f"connect to nats stream: '{self.stream_name}' on subject: '{subject}' "
            f"delayed batch tasks: {list(task_params.keys())}"
        )


async def send_task_by_subject(nc, subject, *args, **kwargs) -> None:
    if not subject:
        raise ValueError("subject is empty")

    meta = MetaTask(
        args=[normalize_payload(v) for v in args],
        kwargs={k: normalize_payload(v) for k, v in kwargs.items()},
    )

    await nc.publish(subject, meta.model_dump(mode="json"))
    logger.info(f"pub task message '{meta.task}' to subject '{subject}' with args={meta.args} kwargs={meta.kwargs}")



import asyncio
from collections.abc import Awaitable
from dataclasses import dataclass
from typing import Any, Callable, Optional

from nats.aio import subscription
from nats.aio.client import Callback
from nats.js import api, client


@dataclass
class SubscriptionMeta:
    subject: str
    queue: str = ""
    handler: Optional[Callable[[Any], Awaitable[None]]] = None
    future: Optional[asyncio.Future] = None
    max_msgs: int = 0
    pending_msgs_limit: int = subscription.DEFAULT_SUB_PENDING_MSGS_LIMIT
    pending_bytes_limit: int = subscription.DEFAULT_SUB_PENDING_BYTES_LIMIT


@dataclass
class PushSubscriptionMeta:
    subject: str
    queue: Optional[str] = None
    cb: Optional[Callback] = None
    durable: Optional[str] = None
    stream: Optional[str] = None
    config: Optional[api.ConsumerConfig] = None
    manual_ack: bool = False
    ordered_consumer: bool = False
    idle_heartbeat: Optional[float] = None
    flow_control: bool = False
    pending_msgs_limit: int = client.DEFAULT_JS_SUB_PENDING_MSGS_LIMIT
    pending_bytes_limit: int = client.DEFAULT_JS_SUB_PENDING_BYTES_LIMIT
    deliver_policy: Optional[api.DeliverPolicy] = None
    headers_only: Optional[bool] = None
    inactive_threshold: Optional[float] = None


@dataclass
class PullSubscriptionMeta:
    handler: Callback
    subject: str
    durable: Optional[str] = None
    stream: Optional[str] = None
    config: Optional[api.ConsumerConfig] = None
    pending_msgs_limit: int = client.DEFAULT_JS_SUB_PENDING_MSGS_LIMIT
    pending_bytes_limit: int = client.DEFAULT_JS_SUB_PENDING_BYTES_LIMIT
    inbox_prefix: bytes = api.INBOX_PREFIX
    # fetch
    batch: int = (1,)
    timeout: Optional[float] = (None,)
    heartbeat: Optional[float] = None
